#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <richedit.h>
#include "AkelDLL.h"


//Defines
#define STRID_DOCNAME     1
#define STRID_ERROR_PRINT 2

//Functions prototypes
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void RichPrintA(HWND hWnd, char *pFile, BOOL bSilent);
void RichPrintW(HWND hWnd, wchar_t *wpFile, BOOL bSilent);
void GetPrinterDCA(PRINTDLGA *pd);
void GetPrinterDCW(PRINTDLGW *pd);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
HWND hMainWnd;
BOOL bOldWindows;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
WNDPROCDATA *NewMainProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(-1, -1, -1, -1);
  pv->pPluginName="PrintClassic";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  //Is plugin already loaded?
  if (bInitMain)
  {
    UninitMain();
  }
  else
  {
    InitMain();

    //Stay in memory, and show as active
    pd->nUnload=UD_NONUNLOAD_ACTIVE;
  }
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDM_FILE_PRINT)
    {
      EDITINFO ei;

      SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei);
      if (bOldWindows)
        RichPrintA(ei.hWndEdit, (char *)ei.pFile, FALSE);
      else
        RichPrintW(ei.hWndEdit, (wchar_t *)ei.pFile, FALSE);
      return FALSE;
    }
  }

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

void RichPrintA(HWND hWnd, char *pFile, BOOL bSilent)
{
  PAGESETUPDLGA *psdg;
  PRINTDLGA *pdg;
  DOCINFOA di={0};
  FORMATRANGE fr;
  RECT rcInitial;
  int nWidth;
  int nHeight;
  int nNextCh;
  BOOL bError=FALSE;

  pdg=(PRINTDLGA *)SendMessage(hMainWnd, AKD_GETPRINTDLG, 0, 0);
  psdg=(PAGESETUPDLGA *)SendMessage(hMainWnd, AKD_GETPAGEDLG, 0, 0);

  if (bSilent)
  {
    GetPrinterDCA(pdg);
    psdg->hDevMode=pdg->hDevMode;
    psdg->hDevNames=pdg->hDevNames;
  }
  else
  {
    if (!PrintDlgA(pdg)) return;
    psdg->hDevMode=pdg->hDevMode;
    psdg->hDevNames=pdg->hDevNames;
  }

  //About 56 twips per millimeter
  nWidth=GetDeviceCaps(pdg->hDC, HORZSIZE) * 56;
  nHeight=GetDeviceCaps(pdg->hDC, VERTSIZE) * 56;
  fr.hdc=fr.hdcTarget=pdg->hDC;

  if (psdg->Flags & PSD_INHUNDREDTHSOFMILLIMETERS)
  {
    fr.rc.left=psdg->rtMargin.left * 567 / 1000;
    fr.rc.top=psdg->rtMargin.top * 567 / 1000;
    fr.rc.right=nWidth - psdg->rtMargin.right * 567 / 1000;
    fr.rc.bottom=nHeight - psdg->rtMargin.bottom * 567 / 1000;
  }
  else
  {
    fr.rc.left=psdg->rtMargin.left * 144 / 100;
    fr.rc.top=psdg->rtMargin.top * 144 / 100;
    fr.rc.right=nWidth - psdg->rtMargin.right * 144 / 100;
    fr.rc.bottom=nHeight - psdg->rtMargin.bottom * 144 / 100;
  }
  fr.rcPage.left=0;
  fr.rcPage.top=0;
  fr.rcPage.right=nWidth;
  fr.rcPage.bottom=nHeight;

  if (pdg->Flags & PD_SELECTION)
  {
    SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&fr.chrg);
  }
  else
  {
    fr.chrg.cpMin=0;
    fr.chrg.cpMax=-1;
  }

  di.cbSize=sizeof(DOCINFOA);
  di.lpszDocName=(*pFile)?pFile:GetLangStringA(wLangSystem, STRID_DOCNAME);
  if (pdg->Flags & PD_PRINTTOFILE) di.lpszOutput="FILE:";

  if (StartDocA(pdg->hDC, &di) > 0)
  {
    rcInitial=fr.rc;

    while (!bError)
    {
      nNextCh=SendMessage(hWnd, EM_FORMATRANGE, FALSE, (LPARAM)&fr);
      if (nNextCh <= fr.chrg.cpMin) break;
      fr.chrg.cpMin=nNextCh;

      StartPage(pdg->hDC);
      if (!SendMessage(hWnd, EM_DISPLAYBAND, 0, (LPARAM)&fr.rc)) bError=TRUE;
      fr.rc=rcInitial;
      EndPage(pdg->hDC);
    }
    SendMessage(hWnd, EM_FORMATRANGE, FALSE, (LPARAM)NULL);
    EndDoc(pdg->hDC);
  }
  else bError=TRUE;

  DeleteDC(pdg->hDC);
  if (bError) MessageBoxA(hMainWnd, GetLangStringA(wLangSystem, STRID_ERROR_PRINT), "PrintClassic plugin", MB_OK|MB_ICONERROR);
}

void RichPrintW(HWND hWnd, wchar_t *wpFile, BOOL bSilent)
{
  PAGESETUPDLGW *psdg;
  PRINTDLGW *pdg;
  DOCINFOW di={0};
  FORMATRANGE fr;
  RECT rcInitial;
  int nWidth;
  int nHeight;
  int nNextCh;
  BOOL bError=FALSE;

  pdg=(PRINTDLGW *)SendMessage(hMainWnd, AKD_GETPRINTDLG, 0, 0);
  psdg=(PAGESETUPDLGW *)SendMessage(hMainWnd, AKD_GETPAGEDLG, 0, 0);

  if (bSilent)
  {
    GetPrinterDCW(pdg);
    psdg->hDevMode=pdg->hDevMode;
    psdg->hDevNames=pdg->hDevNames;
  }
  else
  {
    if (!PrintDlgW(pdg)) return;
    psdg->hDevMode=pdg->hDevMode;
    psdg->hDevNames=pdg->hDevNames;
  }

  //About 56 twips per millimeter
  nWidth=GetDeviceCaps(pdg->hDC, HORZSIZE) * 56;
  nHeight=GetDeviceCaps(pdg->hDC, VERTSIZE) * 56;
  fr.hdc=fr.hdcTarget=pdg->hDC;

  if (psdg->Flags & PSD_INHUNDREDTHSOFMILLIMETERS)
  {
    fr.rc.left=psdg->rtMargin.left * 567 / 1000;
    fr.rc.top=psdg->rtMargin.top * 567 / 1000;
    fr.rc.right=nWidth - psdg->rtMargin.right * 567 / 1000;
    fr.rc.bottom=nHeight - psdg->rtMargin.bottom * 567 / 1000;
  }
  else
  {
    fr.rc.left=psdg->rtMargin.left * 144 / 100;
    fr.rc.top=psdg->rtMargin.top * 144 / 100;
    fr.rc.right=nWidth - psdg->rtMargin.right * 144 / 100;
    fr.rc.bottom=nHeight - psdg->rtMargin.bottom * 144 / 100;
  }
  fr.rcPage.left=0;
  fr.rcPage.top=0;
  fr.rcPage.right=nWidth;
  fr.rcPage.bottom=nHeight;

  if (pdg->Flags & PD_SELECTION)
  {
    SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&fr.chrg);
  }
  else
  {
    fr.chrg.cpMin=0;
    fr.chrg.cpMax=-1;
  }

  di.cbSize=sizeof(DOCINFOW);
  di.lpszDocName=(*wpFile)?wpFile:GetLangStringW(wLangSystem, STRID_DOCNAME);
  if (pdg->Flags & PD_PRINTTOFILE) di.lpszOutput=L"FILE:";

  if (StartDocW(pdg->hDC, &di) > 0)
  {
    rcInitial=fr.rc;

    while (!bError)
    {
      nNextCh=SendMessage(hWnd, EM_FORMATRANGE, FALSE, (LPARAM)&fr);
      if (nNextCh <= fr.chrg.cpMin) break;
      fr.chrg.cpMin=nNextCh;

      StartPage(pdg->hDC);
      if (!SendMessage(hWnd, EM_DISPLAYBAND, 0, (LPARAM)&fr.rc)) bError=TRUE;
      fr.rc=rcInitial;
      EndPage(pdg->hDC);
    }
    SendMessage(hWnd, EM_FORMATRANGE, FALSE, (LPARAM)NULL);
    EndDoc(pdg->hDC);
  }
  else bError=TRUE;

  DeleteDC(pdg->hDC);
  if (bError) MessageBoxW(hMainWnd, GetLangStringW(wLangSystem, STRID_ERROR_PRINT), L"PrintClassic plugin", MB_OK|MB_ICONERROR);
}

void GetPrinterDCA(PRINTDLGA *pd)
{
  PRINTDLGA pdTmp;
  DEVMODEA *hDevMode;
  DEVNAMES *hDevNames;

  pd->hDC=NULL;

  if (pd->hDevMode && pd->hDevNames)
  {
    hDevMode=(DEVMODEA *)GlobalLock(pd->hDevMode);
    hDevNames=(DEVNAMES *)GlobalLock(pd->hDevNames);
    pd->hDC=CreateDCA((char *)hDevNames + hDevNames->wDriverOffset, (char *)hDevNames + hDevNames->wDeviceOffset, NULL, hDevMode);
    GlobalUnlock(pd->hDevMode);
    GlobalUnlock(pd->hDevNames);
  }
  if (!pd->hDC)
  {
    memcpy(&pdTmp, pd, sizeof(PRINTDLGA));
    pdTmp.Flags|=PD_RETURNDEFAULT|PD_RETURNDC;
    pdTmp.hDevMode=NULL;
    pdTmp.hDevNames=NULL;

    if (PrintDlgA(&pdTmp))
    {
     pd->hDC=pdTmp.hDC;
     pd->hDevMode=pdTmp.hDevMode;
     pd->hDevNames=pdTmp.hDevNames;
    }
  }
}

void GetPrinterDCW(PRINTDLGW *pd)
{
  PRINTDLGW pdTmp;
  DEVMODEW *hDevMode;
  DEVNAMES *hDevNames;

  pd->hDC=NULL;

  if (pd->hDevMode && pd->hDevNames)
  {
    hDevMode=(DEVMODEW *)GlobalLock(pd->hDevMode);
    hDevNames=(DEVNAMES *)GlobalLock(pd->hDevNames);
    pd->hDC=CreateDCW((wchar_t *)hDevNames + hDevNames->wDriverOffset, (wchar_t *)hDevNames + hDevNames->wDeviceOffset, NULL, hDevMode);
    GlobalUnlock(pd->hDevMode);
    GlobalUnlock(pd->hDevNames);
  }
  if (!pd->hDC)
  {
    memcpy(&pdTmp, pd, sizeof(PRINTDLGW));
    pdTmp.Flags|=PD_RETURNDEFAULT|PD_RETURNDC;
    pdTmp.hDevMode=NULL;
    pdTmp.hDevNames=NULL;

    if (PrintDlgW(&pdTmp))
    {
     pd->hDC=pdTmp.hDC;
     pd->hDevMode=pdTmp.hDevMode;
     pd->hDevNames=pdTmp.hDevNames;
    }
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_DOCNAME)
      return L"\x0041\x006B\x0065\x006C\x0050\x0061\x0064\x0020\x0434\x043E\x043A\x0443\x043C\x0435\x043D\x0442";
    if (nStringID == STRID_ERROR_PRINT)
      return L"\x041E\x0448\x0438\x0431\x043A\x0430\x0020\x043F\x0440\x0438\x0020\x043F\x0435\x0447\x0430\x0442\x0438\x0020\x0434\x043E\x043A\x0443\x043C\x0435\x043D\x0442\x0430\x002E";
  }
  else
  {
    if (nStringID == STRID_DOCNAME)
      return L"AkelPad document";
    if (nStringID == STRID_ERROR_PRINT)
      return L"Cannot print document.";
  }
  return L"";
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hMainWnd=pd->hMainWnd;
  bOldWindows=pd->bOldWindows;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);
}

void InitMain()
{
  bInitMain=TRUE;

  //SubClass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProc, (LPARAM)&NewMainProcData);
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
